%%%-------------------------------------------------------------------
%%% @author Fabio Salvini <fs@fabiosalvini.com>
%%% @copyright (C) 2017, Fabio Salvini
%%% @doc
%%%
%%% @end
%%% Created :  2 Jul 2017 by Fabio Salvini <fs@fabiosalvini.com>
%%%-------------------------------------------------------------------
-module(gatherer).

-behaviour(gen_server).

%% API
-export([start_link/2]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2,
         terminate/2, code_change/3]).

-define(SERVER, ?MODULE).

-record(log, {file, error_regex}).
%% -record(state, {}).

%%%===================================================================
%%% API
%%%===================================================================

%%--------------------------------------------------------------------
%% @doc
%% Starts the server
%%
%% @spec start_link(File, ErrorRegex) -> {ok, Pid} |
%%                                       ignore |
%%                                       {error, Error}
%% @end
%%--------------------------------------------------------------------
start_link(File, ErrorRegex) ->
    gen_server:start_link(?MODULE, [#log{file = File, error_regex = ErrorRegex}], []).

%%%===================================================================
%%% gen_server callbacks
%%%===================================================================

%%--------------------------------------------------------------------
%% @private
%% @doc
%% Initializes the server
%%
%% @spec init(Args) -> {ok, State} |
%%                     {ok, State, Timeout} |
%%                     ignore |
%%                     {stop, Reason}
%% @end
%%--------------------------------------------------------------------
init([Log]) ->
    {ok, [off, Log]}.

%%--------------------------------------------------------------------
%% @private
%% @doc
%% Handling call messages
%%
%% @spec handle_call(Request, From, State) ->
%%                                   {reply, Reply, State} |
%%                                   {reply, Reply, State, Timeout} |
%%                                   {noreply, State} |
%%                                   {noreply, State, Timeout} |
%%                                   {stop, Reason, Reply, State} |
%%                                   {stop, Reason, State}
%% @end
%%--------------------------------------------------------------------
handle_call(_Request, _From, State) ->
    Reply = ok,
    {reply, Reply, State}.

%%--------------------------------------------------------------------
%% @private
%% @doc
%% Handling cast messages
%%
%% @spec handle_cast(Msg, State) -> {noreply, State} |
%%                                  {noreply, State, Timeout} |
%%                                  {stop, Reason, State}
%% @end
%%--------------------------------------------------------------------
handle_cast(_Msg, State) ->
    {noreply, State}.

%%--------------------------------------------------------------------
%% @private
%% @doc
%% Handling all non call/cast messages
%%
%% @spec handle_info(Info, State) -> {noreply, State} |
%%                                   {noreply, State, Timeout} |
%%                                   {stop, Reason, State}
%% @end
%%--------------------------------------------------------------------
handle_info({log_line, Text}, [off, Log = #log{error_regex = ErrorRegex}]) ->
    case isError(Text, ErrorRegex) of
        true ->
            {ok, Timer} = timer:send_after(1000, {timeout}),
            {ok, SafeTimer} = timer:send_after(30000, {timeout}),
            {noreply, [on, Log, Text, Timer, SafeTimer]};
        false -> {noreply, [off, Log]}
    end;
handle_info({log_line, Text}, [on, Log = #log{error_regex = ErrorRegex}, Error, Timer, SafeTimer]) ->
    case isError(Text, ErrorRegex) of
        true ->
            timer:cancel(Timer),
            {ok, NewTimer} = timer:send_after(1000, {timeout}),
            {noreply, [on, Log, Error ++ Text, NewTimer, SafeTimer]};
        false ->
            {noreply, [on, Log, Error ++ Text, Timer, SafeTimer]}
    end;
handle_info({timeout}, [on, Log = #log{file = File}, Error, Timer, SafeTimer]) ->
    timer:cancel(Timer),
    timer:cancel(SafeTimer),
    mailer ! {error, File, Error},
    {noreply, [off, Log]}.

%%--------------------------------------------------------------------
%% @private
%% @doc
%% This function is called by a gen_server when it is about to
%% terminate. It should be the opposite of Module:init/1 and do any
%% necessary cleaning up. When it returns, the gen_server terminates
%% with Reason. The return value is ignored.
%%
%% @spec terminate(Reason, State) -> void()
%% @end
%%--------------------------------------------------------------------
terminate(_Reason, _State) ->
    shutdown.

%%--------------------------------------------------------------------
%% @private
%% @doc
%% Convert process state when code is changed
%%
%% @spec code_change(OldVsn, State, Extra) -> {ok, NewState}
%% @end
%%--------------------------------------------------------------------
code_change(_OldVsn, State, _Extra) ->
    {ok, State}.

%%%===================================================================
%%% Internal functions
%%%===================================================================
isError(Text, ErrorRegex) ->
    case re:run(Text, ErrorRegex) of
        {match, _} ->
            true;
        nomatch ->
            false
    end.
